import functools
import time

from eth_tester.backends.mock.common import (
    calculate_effective_gas_price,
)
from eth_tester.utils.transactions import (
    extract_transaction_type,
)
from eth_typing import (
    Hash32,
)
from eth_utils import (
    apply_to_return_value,
    is_bytes,
    is_dict,
    is_integer,
    is_list_like,
    is_null,
    is_text,
    keccak,
    to_bytes,
    to_dict,
    to_tuple,
)

from eth_utils.toolz import (
    assoc,
)

from eth_tester.backends.common import merge_genesis_overrides
from eth_tester.constants import DYNAMIC_FEE_TRANSACTION_PARAMS
from eth_tester.utils.address import (
    generate_contract_address,
)


ZERO_32BYTES = b"\x00" * 32
ZERO_8BYTES = b"\x00" * 8
ZERO_ADDRESS = b"\x00" * 20
BLOCK_ELASTICITY_MULTIPLIER = 2
BASE_FEE_MAX_CHANGE_DENOMINATOR = 8


@apply_to_return_value(b"|".join)
@to_tuple
def bytes_repr(value):
    if is_bytes(value):
        yield value
    elif is_text(value):
        yield to_bytes(text=value)
    elif is_list_like(value):
        yield b"".join(
            (
                b"(",
                b",".join(bytes_repr(item) for item in value),
                b")",
            )
        )
    elif is_dict(value):
        yield b"".join(
            (
                b"{",
                b",".join(
                    (
                        b":".join((bytes_repr(key), bytes_repr(item)))
                        for key, item in value.items()
                    )
                ),
                b"}",
            )
        )
    elif is_integer(value):
        yield to_bytes(value)
    elif is_null(value):
        yield "None@{}".format(id(value))
    else:
        raise TypeError("Unsupported type for bytes_repr: {}".format(type(value)))


def fake_rlp_hash(value):
    return keccak(bytes_repr(value))


def add_hash(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        value = fn(*args, **kwargs)
        if "hash" in value:
            return value
        else:
            return assoc(value, "hash", keccak(bytes_repr(value)))

    return inner


def create_transaction(
    transaction, block, transaction_index, is_pending, overrides=None
):
    filled_txn = _fill_transaction(
        transaction, block, transaction_index, is_pending, overrides
    )
    if "hash" in filled_txn:
        return filled_txn
    else:
        return assoc(filled_txn, "hash", fake_rlp_hash(filled_txn))


@to_dict
def _fill_transaction(
    transaction, block, transaction_index, is_pending, overrides=None
):
    is_dynamic_fee_transaction = any(
        _ in transaction for _ in DYNAMIC_FEE_TRANSACTION_PARAMS
    ) or not any(
        _ in transaction for _ in DYNAMIC_FEE_TRANSACTION_PARAMS + ("gas_price",)
    )

    if overrides is None:
        overrides = {}

    if "hash" in overrides:  # else calculate hash after all fields are filled
        yield "hash", overrides["hash"]

    # Here, we yield the key with the overrides value if it exists, else either
    # the transaction value if it exists or a default value
    yield "nonce", overrides.get("nonce", 0)
    yield "from", overrides.get("from", transaction.get("from"))
    yield "to", overrides.get("to", transaction.get("to", b""))
    yield "data", overrides.get("data", transaction.get("data", b""))
    yield "value", overrides.get("value", transaction.get("value", 0))
    yield "gas", overrides.get("gas", transaction.get("gas"))
    yield "r", overrides.get("r", transaction.get("r", 12345))
    yield "s", overrides.get("s", transaction.get("s", 67890))
    yield "v", overrides.get("v", transaction.get("v", 0))

    if is_dynamic_fee_transaction:
        # dynamic fee transaction (type = 2)
        yield "max_fee_per_gas", overrides.get(
            "max_fee_per_gas", transaction.get("max_fee_per_gas", 1000000000)
        )
        yield "max_priority_fee_per_gas", overrides.get(
            "max_priority_fee_per_gas",
            transaction.get("max_priority_fee_per_gas", 1000000000),
        )
        yield from _yield_typed_transaction_fields(overrides, transaction)

    else:
        yield "gas_price", overrides.get("gas_price", transaction.get("gas_price"))
        if "access_list" in transaction:
            # access list transaction (type = 1)
            yield from _yield_typed_transaction_fields(overrides, transaction)


def _yield_typed_transaction_fields(overrides, transaction):
    yield "chain_id", overrides.get(
        "chain_id", transaction.get("chain_id", 131277322940537)
    )
    yield "access_list", overrides.get(
        "access_list", transaction.get("access_list", ())
    )


@to_dict
def make_log(transaction, block, transaction_index, log_index, overrides=None):
    if overrides is None:
        overrides = {}

    is_pending = transaction["block_number"] is None

    if "type" in overrides:
        yield "type", overrides["type"]
    else:
        yield "type", "pending" if is_pending else "mined"

    if "transaction_index" in overrides:
        yield "transaction_index", overrides["transaction_index"]
    else:
        yield "transaction_index", None if is_pending else transaction_index

    if "block_number" in overrides:
        yield "block_number", overrides["block_number"]
    else:
        yield "block_number", None if is_pending else block["number"]

    if "block_hash" in overrides:
        yield "block_hash", overrides["block_hash"]
    else:
        yield "block_hash", None if is_pending else block["hash"]

    if "log_index" in overrides:
        yield "log_index", overrides["log_index"]
    else:
        yield "log_index", log_index

    if "address" in overrides:
        yield "address", overrides["address"]
    else:
        yield "address", transaction.get("to", b"")

    if "data" in overrides:
        yield "data", overrides["data"]
    else:
        yield "data", b""

    if "topics" in overrides:
        yield "topics", overrides["topics"]
    else:
        yield "topics", []


@to_dict
def make_receipt(transaction, block, _transaction_index, overrides=None):
    if overrides is None:
        overrides = {}

    gas_used = overrides.get("gas_used", 21000)
    yield "gas_used", gas_used
    yield "logs", overrides.get("logs", [])
    yield "transaction_hash", overrides.get("transaction_hash", transaction.get("hash"))
    yield (
        "cumulative_gas_used",
        overrides.get("cumulative_gas_used", block.get("gas_used") + gas_used),
    )
    yield (
        "effective_gas_price",
        overrides.get(
            "effective_gas_price", calculate_effective_gas_price(transaction, block)
        ),
    )
    yield (
        "type",
        overrides.get(
            "type", transaction.get("type", extract_transaction_type(transaction))
        ),
    )
    yield (
        "contract_address",
        overrides.get(
            "contract_address",
            generate_contract_address(transaction["from"], transaction["nonce"]),
        ),
    )


GENESIS_NONCE = b"\x00\x00\x00\x00\x00\x00\x00*"  # 42 encoded as big-endian-integer
BLANK_ROOT_HASH = b"V\xe8\x1f\x17\x1b\xccU\xa6\xff\x83E\xe6\x92\xc0\xf8n\x5bH\xe0\x1b\x99l\xad\xc0\x01b/\xb5\xe3c\xb4!"  # noqa: E501
EMPTY_UNCLE_HASH = b"\x1d\xccM\xe8\xde\xc7]z\xab\x85\xb5g\xb6\xcc\xd4\x1a\xd3\x12E\x1b\x94\x8at\x13\xf0\xa1B\xfd@\xd4\x93G"  # noqa: E501
POST_MERGE_DIFFICULTY = 0
POST_MERGE_MIX_HASH = Hash32(32 * b"\x00")
POST_MERGE_NONCE = b"\x00\x00\x00\x00\x00\x00\x00\x00"


def make_genesis_block(overrides=None):
    default_genesis_block = {
        "number": 0,
        "hash": ZERO_32BYTES,
        "parent_hash": ZERO_32BYTES,
        "nonce": POST_MERGE_NONCE,
        "sha3_uncles": EMPTY_UNCLE_HASH,
        "logs_bloom": 0,
        "transactions_root": BLANK_ROOT_HASH,
        "receipts_root": BLANK_ROOT_HASH,
        "state_root": BLANK_ROOT_HASH,
        "coinbase": ZERO_ADDRESS,
        "difficulty": POST_MERGE_DIFFICULTY,
        "mix_hash": POST_MERGE_MIX_HASH,
        "total_difficulty": 131072,
        "size": 0,
        "extra_data": ZERO_32BYTES,
        # gas limit at London fork block 12965000 on mainnet
        "gas_limit": 30029122,
        "gas_used": 0,
        "timestamp": int(time.time()),
        "transactions": [],
        "uncles": [],
        # base fee at London fork block 12965000 on mainnet
        "base_fee_per_gas": 1000000000,
        "withdrawals": [],
        "withdrawals_root": BLANK_ROOT_HASH,
    }
    if overrides is not None:
        genesis_block = merge_genesis_overrides(
            defaults=default_genesis_block, overrides=overrides
        )
    else:
        genesis_block = default_genesis_block
    return genesis_block


@add_hash
@to_dict
def make_block_from_parent(parent_block, overrides=None):
    if overrides is None:
        overrides = {}

    if "number" in overrides:
        yield "number", overrides["number"]
    else:
        yield "number", parent_block["number"] + 1

    if "hash" in overrides:
        yield "hash", overrides["hash"]
    else:
        yield "hash", keccak(parent_block["hash"])

    if "parent_hash" in overrides:
        yield "parent_hash", overrides["parent_hash"]
    else:
        yield "parent_hash", parent_block["hash"]

    if "nonce" in overrides:
        yield "nonce", overrides["nonce"]
    else:
        yield "nonce", parent_block["nonce"]

    if "sha3_uncles" in overrides:
        yield "sha3_uncles", overrides["sha3_uncles"]
    else:
        yield "sha3_uncles", EMPTY_UNCLE_HASH

    if "logs_bloom" in overrides:
        yield "logs_bloom", overrides["logs_bloom"]
    else:
        yield "logs_bloom", 0

    if "transactions_root" in overrides:
        yield "transactions_root", overrides["transactions_root"]
    else:
        yield "transactions_root", BLANK_ROOT_HASH

    if "receipts_root" in overrides:
        yield "receipts_root", overrides["receipts_root"]
    else:
        yield "receipts_root", BLANK_ROOT_HASH

    if "state_root" in overrides:
        yield "state_root", overrides["state_root"]
    else:
        yield "state_root", BLANK_ROOT_HASH

    if "coinbase" in overrides:
        yield "coinbase", overrides["coinbase"]
    else:
        yield "coinbase", ZERO_ADDRESS

    if "difficulty" in overrides:
        difficulty = overrides["difficulty"]
    else:
        difficulty = POST_MERGE_DIFFICULTY
    yield "difficulty", difficulty

    if "mix_hash" in overrides:
        yield "mix_hash", overrides["mix_hash"]
    else:
        yield "mix_hash", POST_MERGE_MIX_HASH

    if "total_difficulty" in overrides:
        yield "total_difficulty", overrides["total_difficulty"]
    else:
        yield "total_difficulty", parent_block["difficulty"] + difficulty

    if "size" in overrides:
        yield "size", overrides["size"]
    else:
        yield "size", 0

    if "extra_data" in overrides:
        yield "extra_data", overrides["extra_data"]
    else:
        yield "extra_data", ZERO_32BYTES

    if "gas_limit" in overrides:
        yield "gas_limit", overrides["gas_limit"]
    else:
        yield "gas_limit", parent_block["gas_limit"]

    if "gas_used" in overrides:
        yield "gas_used", overrides["gas_used"]
    else:
        yield "gas_used", 0

    if "timestamp" in overrides:
        yield "timestamp", overrides["timestamp"]
    else:
        yield "timestamp", parent_block["timestamp"] + 1

    if "transactions" in overrides:
        yield "transactions", overrides["transactions"]
    else:
        yield "transactions", []

    if "uncles" in overrides:
        yield "uncles", overrides["uncles"]
    else:
        yield "uncles", []

    if "base_fee_per_gas" in overrides:
        yield "base_fee_per_gas", overrides["base_fee_per_gas"]
    else:
        yield "base_fee_per_gas", _calculate_expected_base_fee_per_gas(parent_block)

    if "withdrawals" in overrides:
        yield "withdrawals", overrides["withdrawals"]
    else:
        yield "withdrawals", []

    if "withdrawals_root" in overrides:
        yield "withdrawals_root", overrides["withdrawals_root"]
    else:
        yield "withdrawals_root", BLANK_ROOT_HASH


def _calculate_expected_base_fee_per_gas(parent_block) -> int:
    """py-evm logic for calculating the base fee from parent header"""
    parent_base_fee_per_gas = parent_block["base_fee_per_gas"]

    parent_gas_target = parent_block["gas_limit"] // BLOCK_ELASTICITY_MULTIPLIER
    parent_gas_used = parent_block["gas_used"]

    if parent_gas_used == parent_gas_target:
        return parent_base_fee_per_gas

    elif parent_gas_used > parent_gas_target:
        gas_used_delta = parent_gas_used - parent_gas_target
        overburnt_wei = parent_base_fee_per_gas * gas_used_delta
        base_fee_per_gas_delta = max(
            overburnt_wei // parent_gas_target // BASE_FEE_MAX_CHANGE_DENOMINATOR,
            1,
        )
        return parent_base_fee_per_gas + base_fee_per_gas_delta

    else:
        gas_used_delta = parent_gas_target - parent_gas_used
        underburnt_wei = parent_base_fee_per_gas * gas_used_delta
        base_fee_per_gas_delta = (
            underburnt_wei // parent_gas_target // BASE_FEE_MAX_CHANGE_DENOMINATOR
        )
        return max(parent_base_fee_per_gas - base_fee_per_gas_delta, 0)



from __future__ import absolute_import

import os
import time
from typing import Dict, List, Union

from eth_abi import abi
from eth_abi.exceptions import DecodingError

from eth_account.hdaccount import HDPath, seed_from_mnemonic
from eth_typing import Address

from eth_utils import (
    encode_hex,
    int_to_big_endian,
    to_bytes,
    to_checksum_address,
    to_dict,
    to_tuple,
    to_wei,
    is_integer,
)

from eth_utils.decorators import replace_exceptions

from eth_utils.toolz import (
    assoc,
)

from eth_keys import KeyAPI

from eth_tester.constants import DYNAMIC_FEE_TRANSACTION_PARAMS
from eth_tester.exceptions import (
    BackendDistributionNotFound,
    BlockNotFound,
    TransactionFailed,
    TransactionNotFound,
    ValidationError,
)

from eth_tester.backends.base import BaseChainBackend
from eth_tester.backends.common import merge_genesis_overrides

from .serializers import (
    serialize_block,
    serialize_transaction,
    serialize_transaction_receipt,
)
from .utils import is_supported_pyevm_version_available
from ...validation.inbound import validate_inbound_withdrawals


if is_supported_pyevm_version_available():
    from eth.constants import (
        GENESIS_PARENT_HASH,
        POST_MERGE_DIFFICULTY,
        POST_MERGE_MIX_HASH,
        POST_MERGE_NONCE,
    )
    from eth.exceptions import (
        HeaderNotFound as EVMHeaderNotFound,
        InvalidInstruction as EVMInvalidInstruction,
        Revert as EVMRevert,
    )
    from eth.vm.forks import (
        ParisVM,
        ShanghaiVM,
    )
    from eth.vm.spoof import SpoofTransaction as EVMSpoofTransaction
    from eth.vm.forks.shanghai.withdrawals import Withdrawal
else:
    EVMHeaderNotFound = None
    EVMInvalidInstruction = None
    EVMRevert = None
    GENESIS_PARENT_HASH = None
    ParisVM = None
    POST_MERGE_DIFFICULTY = None
    POST_MERGE_MIX_HASH = None
    POST_MERGE_NONCE = None
    ShanghaiVM = None
    Withdrawal = None


ZERO_ADDRESS = 20 * b"\x00"
ZERO_HASH32 = 32 * b"\x00"
EIP838_SIG = b"\x08\xc3y\xa0"

EMPTY_RLP_LIST_HASH = b"\x1d\xccM\xe8\xde\xc7]z\xab\x85\xb5g\xb6\xcc\xd4\x1a\xd3\x12E\x1b\x94\x8at\x13\xf0\xa1B\xfd@\xd4\x93G"  # noqa: E501
BLANK_ROOT_HASH = b"V\xe8\x1f\x17\x1b\xccU\xa6\xff\x83E\xe6\x92\xc0\xf8n\x5bH\xe0\x1b\x99l\xad\xc0\x01b/\xb5\xe3c\xb4!"  # noqa: E501


GENESIS_BLOCK_NUMBER = 0
GENESIS_DIFFICULTY = 131072
GENESIS_GAS_LIMIT = 30029122  # gas limit at London fork block 12965000 on mainnet
GENESIS_COINBASE = ZERO_ADDRESS
GENESIS_NONCE = b"\x00\x00\x00\x00\x00\x00\x00*"  # 42 encoded as big-endian-integer
GENESIS_MIX_HASH = ZERO_HASH32
GENESIS_EXTRA_DATA = b""
GENESIS_INITIAL_ALLOC = {}


MINIMUM_GAS_ESTIMATE = 30000
# A buffer of 1.1 would mean allocate 10% more gas than estimated
GAS_ESTIMATE_BUFFER = 1.5


def get_default_account_state(overrides=None):
    default_account_state = {
        "balance": to_wei(1000000, "ether"),
        "storage": {},
        "code": b"",
        "nonce": 0,
    }
    if overrides is not None:
        account_state = merge_genesis_overrides(
            defaults=default_account_state, overrides=overrides
        )
    else:
        account_state = default_account_state
    return account_state


@to_tuple
def get_default_account_keys(quantity=None):
    keys = KeyAPI()
    quantity = quantity or 10
    for i in range(1, quantity + 1):
        pk_bytes = int_to_big_endian(i).rjust(32, b"\x00")
        private_key = keys.PrivateKey(pk_bytes)
        yield private_key


@to_tuple
def get_account_keys_from_mnemonic(mnemonic, quantity=None, hd_path=None):
    keys = KeyAPI()
    seed = seed_from_mnemonic(mnemonic, "")
    quantity = quantity or 10

    if hd_path is None:
        # default HD path
        hd_path = "m/44'/60'/0'"

    for i in range(0, quantity):
        # create unique HDPath to derive the private key for each account
        key = HDPath(f"{hd_path}/{i}").derive(seed)
        private_key = keys.PrivateKey(key)
        yield private_key


@to_dict
def generate_genesis_state_for_keys(account_keys, overrides=None):
    for private_key in account_keys:
        account_state = get_default_account_state(overrides=overrides)
        yield private_key.public_key.to_canonical_address(), account_state


def get_default_genesis_params(overrides=None):
    # Commented out params became un-configurable in py-evm during London refactor.
    # Post-merge now. Set genesis params to expect post-merge validation. If PoS field
    # value defaults are not desired, use the overrides option.
    default_genesis_params = {
        # "bloom": 0,
        "coinbase": GENESIS_COINBASE,
        "difficulty": POST_MERGE_DIFFICULTY,
        "extra_data": GENESIS_EXTRA_DATA,
        "gas_limit": GENESIS_GAS_LIMIT,
        # "gas_used": 0,
        "mix_hash": POST_MERGE_MIX_HASH,
        "nonce": POST_MERGE_NONCE,
        # "block_number": GENESIS_BLOCK_NUMBER,
        # "parent_hash": GENESIS_PARENT_HASH,
        "receipt_root": BLANK_ROOT_HASH,
        "timestamp": int(time.time()),
        "transaction_root": BLANK_ROOT_HASH,
        # "uncles_hash": EMPTY_RLP_LIST_HASH,
    }
    if overrides is not None:
        genesis_params = merge_genesis_overrides(
            default_genesis_params, overrides=overrides
        )
    else:
        genesis_params = default_genesis_params
    return genesis_params


def setup_tester_chain(
    genesis_params=None,
    genesis_state=None,
    num_accounts=None,
    vm_configuration=None,
    mnemonic=None,
    hd_path=None,
    genesis_is_post_merge=True,
):
    from eth.chains.base import MiningChain
    from eth.consensus import (
        NoProofConsensus,
        ConsensusApplier,
    )
    from eth.db import get_db_backend

    if vm_configuration is None:
        vm_config = ((0, ShanghaiVM),)
    else:
        if len(vm_configuration) > 0:
            _genesis_block_num, genesis_vm = vm_configuration[0]
            if not issubclass(genesis_vm, ParisVM):
                genesis_is_post_merge = False
        consensus_applier = ConsensusApplier(NoProofConsensus)
        vm_config = consensus_applier.amend_vm_configuration(vm_configuration)

    class MainnetTesterPosChain(MiningChain):
        # TODO: Once the logic within `MiningChain` is refactored more generally in
        #  py-evm, change this class inheritance to reflect that since a `PosConsensus`
        #  chain does not mine.
        chain_id = 131277322940537
        vm_configuration = vm_config

        def create_header_from_parent(self, parent_header, **header_params):
            # Keep the gas limit constant
            return super().create_header_from_parent(
                parent_header,
                **assoc(header_params, "gas_limit", parent_header.gas_limit),
            )

        def get_transaction_builder(self):
            return super().get_vm().get_transaction_builder()

    if genesis_params is None:
        overrides = {}
        if not genesis_is_post_merge:
            overrides["difficulty"] = GENESIS_DIFFICULTY
            overrides["nonce"] = GENESIS_NONCE
            overrides["mix_hash"] = GENESIS_MIX_HASH
        genesis_params = get_default_genesis_params(overrides=overrides)

    if genesis_state:
        num_accounts = len(genesis_state)

    if mnemonic:
        account_keys = get_account_keys_from_mnemonic(
            mnemonic, quantity=num_accounts, hd_path=hd_path
        )
    else:
        account_keys = get_default_account_keys(quantity=num_accounts)

    if genesis_state is None:
        genesis_state = generate_genesis_state_for_keys(account_keys)

    base_db = get_db_backend()

    chain = MainnetTesterPosChain.from_genesis(base_db, genesis_params, genesis_state)
    return account_keys, chain


def _get_block_by_number(chain, block_number):
    if block_number in ("latest", "safe", "finalized"):
        head_block = chain.get_block()
        return chain.get_canonical_block_by_number(max(0, head_block.number - 1))
    elif block_number == "earliest":
        return chain.get_canonical_block_by_number(0)
    elif block_number == "pending":
        return chain.get_block()
    elif is_integer(block_number):
        # Note: The head block is the pending block. If a block number is passed
        # explicitly here, return the block only if it is already part of the chain
        # (i.e. not pending).
        head_block = chain.get_block()
        if block_number < head_block.number:
            return chain.get_canonical_block_by_number(block_number)

    # fallback
    raise BlockNotFound(f"No block found for block number: {block_number}")


def _get_block_by_hash(chain, block_hash):
    block = chain.get_block_by_hash(block_hash)

    if block.number >= chain.get_block().number:
        raise BlockNotFound(f"No block found for block hash: {block_hash}")

    block_at_height = chain.get_canonical_block_by_number(block.number)
    if block != block_at_height:
        raise BlockNotFound(f"No block found for block hash: {block_hash}")

    return block


def _get_transaction_by_hash(chain, transaction_hash):
    head_block = chain.get_block()
    for index, transaction in enumerate(head_block.transactions):
        if transaction.hash == transaction_hash:
            return head_block, transaction, index
    for block_number in range(head_block.number - 1, -1, -1):
        # TODO: the chain should be able to look these up directly by hash...
        block = chain.get_canonical_block_by_number(block_number)
        for index, transaction in enumerate(block.transactions):
            if transaction.hash == transaction_hash:
                return block, transaction, index
    else:
        raise TransactionNotFound(
            "No transaction found for transaction hash: {}".format(
                encode_hex(transaction_hash)
            )
        )


def _execute_and_revert_transaction(chain, transaction, block_number="latest"):
    vm = _get_vm_for_block_number(chain, block_number)

    state = vm.state
    snapshot = state.snapshot()
    computation = state.apply_transaction(transaction)
    state.revert(snapshot)
    return computation


def _get_vm_for_block_number(chain, block_number):
    block = _get_block_by_number(chain, block_number)
    vm = chain.get_vm(at_header=block.header)
    return vm


class PyEVMBackend(BaseChainBackend):
    chain = None

    def __init__(
        self,
        genesis_parameters=None,
        genesis_state=None,
        vm_configuration=None,
        mnemonic=None,
        hd_path=None,
    ):
        """
        :param genesis_parameters: A dict of chain parameters for overriding default
             values when setting up the chain.
        :param genesis_state: A dict (or list of tuples) matching accounts to state
             properties, such as `balance`.
        :param vm_configuration: The tuple of virtual machines defining a chain
            schedule as used in py-evm's :attr:`eth.chains.base.Chain.vm_configuration`.
            (at author time, a series of block numbers and virtual machines)
        :param mnemonic: A mnemonic str to use when generating accounts.
        """
        if not is_supported_pyevm_version_available():
            raise BackendDistributionNotFound(
                "The `py-evm` package is not available or not up to date. "
                "The `PyEVMBackend` requires py-evm to be installed and importable. "
                "Please install or update the `py-evm` library."
            )

        self.account_keys = None  # set below
        accounts = len(genesis_state) if genesis_state else None
        self.reset_to_genesis(
            genesis_parameters,
            genesis_state,
            accounts,
            vm_configuration,
            mnemonic,
            hd_path,
        )

    @classmethod
    def from_mnemonic(
        cls,
        mnemonic,
        genesis_state_overrides=None,
        num_accounts=None,
        genesis_parameters=None,
        vm_configuration=None,
        hd_path=None,
    ):
        """
        Create a genesis state pre-populated with accounts. A number of accounts can be
        initialized with a mnemonic phrase and heirarchical deterministic path. If no
        overrides are provided, a default set of accounts will be used.
        """
        genesis_state = PyEVMBackend.generate_genesis_state(
            mnemonic=mnemonic,
            overrides=genesis_state_overrides or {},
            num_accounts=num_accounts,
            hd_path=hd_path,
        )

        return cls(
            genesis_parameters=genesis_parameters,
            genesis_state=genesis_state,
            vm_configuration=vm_configuration,
            mnemonic=mnemonic,
            hd_path=hd_path,
        )

    #
    # Genesis
    #

    @classmethod
    def generate_genesis_params(cls, overrides=None):
        return cls._generate_genesis_params(overrides=overrides)

    @staticmethod
    def _generate_genesis_params(overrides=None):
        return get_default_genesis_params(overrides=overrides)

    @classmethod
    def generate_genesis_state(
        cls, overrides=None, num_accounts=None, mnemonic=None, hd_path=None
    ):
        return cls._generate_genesis_state(
            overrides=overrides,
            num_accounts=num_accounts,
            mnemonic=mnemonic,
            hd_path=hd_path,
        )

    @staticmethod
    def _generate_genesis_state(
        overrides=None, num_accounts=None, mnemonic=None, hd_path=None
    ):
        if mnemonic:
            account_keys = get_account_keys_from_mnemonic(
                mnemonic, quantity=num_accounts, hd_path=hd_path
            )
        else:
            account_keys = get_default_account_keys(quantity=num_accounts)

        return generate_genesis_state_for_keys(
            account_keys=account_keys, overrides=overrides
        )

    def reset_to_genesis(
        self,
        genesis_params=None,
        genesis_state=None,
        num_accounts=None,
        vm_configuration=None,
        mnemonic=None,
        hd_path=None,
    ):
        self.account_keys, self.chain = setup_tester_chain(
            genesis_params,
            genesis_state,
            num_accounts,
            vm_configuration,
            mnemonic,
            hd_path,
        )

    #
    # Private Accounts API
    #
    @property
    def _key_lookup(self):
        return {key.public_key.to_canonical_address(): key for key in self.account_keys}

    #
    # Snapshot API
    #
    def take_snapshot(self):
        block = _get_block_by_number(self.chain, "latest")
        return block.hash

    def revert_to_snapshot(self, snapshot):
        block = self.chain.get_block_by_hash(snapshot)
        chaindb = self.chain.chaindb

        chaindb._set_as_canonical_chain_head(
            chaindb.db, block.header, GENESIS_PARENT_HASH
        )
        if block.number > 0:
            self.chain.import_block(block)
        else:
            self.chain = self.chain.from_genesis_header(chaindb.db, block.header)

    #
    # Meta
    #
    def time_travel(self, to_timestamp):
        # timestamp adjusted by 1 b/c a second is added in mine_blocks
        self.chain.header = self.chain.header.copy(timestamp=(to_timestamp - 1))
        self.mine_blocks()
        return to_timestamp

    #
    # Importing blocks
    #
    @to_tuple
    def mine_blocks(self, num_blocks=1, coinbase=ZERO_ADDRESS):
        mine_kwargs = {"coinbase": coinbase}

        for _ in range(num_blocks):
            if isinstance(self.chain.get_vm(), ParisVM):
                # post-merge, generate a random `mix_hash` to simulate the
                # `prevrandao` value.
                mine_kwargs["mix_hash"] = os.urandom(32)

            block = self.chain.mine_block(**mine_kwargs)
            yield block.hash

    #
    # Accounts
    #
    @to_tuple
    def get_accounts(self):
        for private_key in self.account_keys:
            yield private_key.public_key.to_canonical_address()

    def add_account(self, private_key):
        keys = KeyAPI()
        self.account_keys = self.account_keys + (keys.PrivateKey(private_key),)

    #
    # Chain data
    #
    @replace_exceptions(
        {
            EVMHeaderNotFound: BlockNotFound,
        }
    )
    def get_block_by_number(self, block_number, full_transaction=True):
        block = _get_block_by_number(self.chain, block_number)
        is_pending = block.number == self.chain.get_block().number
        return serialize_block(block, full_transaction, is_pending)

    @replace_exceptions(
        {
            EVMHeaderNotFound: BlockNotFound,
        }
    )
    def get_block_by_hash(self, block_hash, full_transaction=True):
        block = _get_block_by_hash(self.chain, block_hash)
        is_pending = block.number == self.chain.get_block().number
        return serialize_block(block, full_transaction, is_pending)

    def get_transaction_by_hash(self, transaction_hash):
        block, transaction, transaction_index = _get_transaction_by_hash(
            self.chain,
            transaction_hash,
        )
        is_pending = block.number == self.chain.get_block().number
        return serialize_transaction(block, transaction, transaction_index, is_pending)

    def get_transaction_receipt(self, transaction_hash):
        block, transaction, transaction_index = _get_transaction_by_hash(
            self.chain,
            transaction_hash,
        )
        is_pending = block.number == self.chain.get_block().number
        block_receipts = block.get_receipts(self.chain.chaindb)

        return serialize_transaction_receipt(
            block,
            block_receipts,
            transaction,
            transaction_index,
            is_pending,
        )

    #
    # Account state
    #
    def get_nonce(self, account, block_number="latest"):
        vm = _get_vm_for_block_number(self.chain, block_number)
        return vm.state.get_nonce(account)

    def get_balance(self, account, block_number="latest"):
        vm = _get_vm_for_block_number(self.chain, block_number)
        return vm.state.get_balance(account)

    def get_code(self, account, block_number="latest"):
        vm = _get_vm_for_block_number(self.chain, block_number)
        return vm.state.get_code(account)

    def get_storage(self, account: Address, slot: int, block_number="latest") -> bytes:
        vm = _get_vm_for_block_number(self.chain, block_number)
        return vm.state.get_storage(account, slot)

    def get_base_fee(self, block_number="latest"):
        vm = _get_vm_for_block_number(self.chain, block_number)
        return vm.state.base_fee

    #
    # Transactions
    #
    @to_dict
    def _normalize_transaction(self, transaction, block_number="latest"):
        is_dynamic_fee_transaction = (
            any(_ in transaction for _ in DYNAMIC_FEE_TRANSACTION_PARAMS)
            or
            # if no fee params exist, default to dynamic fee transaction:
            not any(
                _ in transaction
                for _ in DYNAMIC_FEE_TRANSACTION_PARAMS + ("gas_price",)
            )
        )
        is_typed_transaction = (
            is_dynamic_fee_transaction or "access_list" in transaction
        )

        for key in transaction:
            if key in ("from", "type"):
                continue
            if key == "v" and is_typed_transaction:
                yield "y_parity", transaction[
                    "v"
                ]  # use y_parity for typed txns, internally
                continue
            yield key, transaction[key]

        if "nonce" not in transaction:
            yield "nonce", self.get_nonce(transaction["from"], block_number)
        if "data" not in transaction:
            yield "data", b""
        if "value" not in transaction:
            yield "value", 0
        if "to" not in transaction:
            yield "to", b""

        if is_dynamic_fee_transaction:
            if not any(_ in transaction for _ in DYNAMIC_FEE_TRANSACTION_PARAMS):
                yield "max_fee_per_gas", 1 * 10**9
                yield "max_priority_fee_per_gas", 1 * 10**9
            elif (
                "max_priority_fee_per_gas" in transaction
                and "max_fee_per_gas" not in transaction
            ):
                yield (
                    "max_fee_per_gas",
                    transaction["max_priority_fee_per_gas"]
                    + 2 * self.get_base_fee(block_number),
                )

        if is_typed_transaction:
            # typed transaction
            if "access_list" not in transaction:
                yield "access_list", ()
            if "chain_id" not in transaction:
                yield "chain_id", self.chain.chain_id

    def _get_normalized_and_unsigned_evm_transaction(
        self, transaction, block_number="latest"
    ):
        normalized_transaction = self._normalize_transaction(transaction, block_number)
        evm_transaction = self._create_type_aware_unsigned_transaction(
            normalized_transaction
        )
        return evm_transaction

    def _get_normalized_and_signed_evm_transaction(
        self, transaction, block_number="latest"
    ):
        if transaction["from"] not in self._key_lookup:
            raise ValidationError(
                'No valid "from" key was provided in the transaction '
                "which is required for transaction signing."
            )
        signing_key = self._key_lookup[transaction["from"]]
        normalized_transaction = self._normalize_transaction(transaction, block_number)
        evm_transaction = self._create_type_aware_unsigned_transaction(
            normalized_transaction
        )
        return evm_transaction.as_signed_transaction(signing_key)

    def _create_type_aware_unsigned_transaction(self, normalized_txn):
        if all(_ in normalized_txn for _ in ("access_list", "gas_price")):
            return self.chain.get_transaction_builder().new_unsigned_access_list_transaction(  # noqa: E501
                **normalized_txn
            )
        elif all(_ in normalized_txn for _ in DYNAMIC_FEE_TRANSACTION_PARAMS):
            return self.chain.get_transaction_builder().new_unsigned_dynamic_fee_transaction(  # noqa: E501
                **normalized_txn
            )
        return self.chain.create_unsigned_transaction(**normalized_txn)

    def send_raw_transaction(self, raw_transaction):
        vm = _get_vm_for_block_number(self.chain, "latest")
        evm_transaction = vm.get_transaction_builder().decode(raw_transaction)
        self.chain.apply_transaction(evm_transaction)
        return evm_transaction.hash

    def send_signed_transaction(self, signed_transaction, block_number="latest"):
        normalized_transaction = self._normalize_transaction(
            signed_transaction, block_number
        )
        signed_evm_transaction = self._create_type_aware_signed_transaction(
            normalized_transaction
        )
        self.chain.apply_transaction(signed_evm_transaction)
        return signed_evm_transaction.hash

    def _create_type_aware_signed_transaction(self, normalized_txn):
        if all(_ in normalized_txn for _ in ("access_list", "gas_price")):
            return self.chain.get_transaction_builder().new_access_list_transaction(
                **normalized_txn
            )
        elif all(_ in normalized_txn for _ in DYNAMIC_FEE_TRANSACTION_PARAMS):
            return self.chain.get_transaction_builder().new_dynamic_fee_transaction(
                **normalized_txn
            )
        return self.chain.create_transaction(**normalized_txn)

    def send_transaction(self, transaction):
        signed_evm_transaction = self._get_normalized_and_signed_evm_transaction(
            transaction,
        )
        self.chain.apply_transaction(signed_evm_transaction)
        return signed_evm_transaction.hash

    def apply_withdrawals(
        self,
        withdrawals_list: List[Dict[str, Union[int, str]]],
    ) -> None:
        """
        Apply withdrawals to the state and mine the block that includes the withdrawals.
        """
        validate_inbound_withdrawals(withdrawals_list)

        vm = _get_vm_for_block_number(self.chain, "latest")
        if not isinstance(vm, ShanghaiVM):
            raise ValidationError(
                "Withdrawals are only supported after the Shanghai fork"
            )

        withdrawals = [
            Withdrawal(
                index=withdrawal_dict["index"],
                validator_index=withdrawal_dict["validator_index"],
                address=to_bytes(
                    hexstr=to_checksum_address(withdrawal_dict["address"])
                ),
                amount=withdrawal_dict["amount"],
            )
            for withdrawal_dict in withdrawals_list
        ]
        self.chain.mine_all(transactions=[], withdrawals=withdrawals)

    def _max_available_gas(self):
        header = self.chain.get_block().header
        return header.gas_limit - header.gas_used

    @replace_exceptions(
        {EVMInvalidInstruction: TransactionFailed, EVMRevert: TransactionFailed}
    )
    def estimate_gas(self, transaction, block_number="latest"):
        evm_transaction = self._get_normalized_and_unsigned_evm_transaction(
            assoc(transaction, "gas", 21000), block_number
        )
        spoofed_transaction = EVMSpoofTransaction(
            evm_transaction, from_=transaction["from"]
        )

        if block_number in ("latest", "safe", "finalized"):
            return self.chain.estimate_gas(spoofed_transaction)
        elif block_number == "earliest":
            return self.chain.estimate_gas(
                spoofed_transaction, self.chain.get_canonical_block_header_by_number(0)
            )
        elif block_number == "pending":
            raise NotImplementedError(
                '"pending" block identifier is unsupported in eth-tester'
            )
        else:
            return self.chain.estimate_gas(
                spoofed_transaction,
                self.chain.get_canonical_block_header_by_number(block_number),
            )

    def is_eip838_error(self, error):
        if not isinstance(error, EVMRevert):
            return False
        elif len(error.args) == 0:
            return False

        try:
            return error.args[0][:4] == EIP838_SIG
        except TypeError:
            return False

    def call(self, transaction, block_number="latest"):
        # TODO: move this to the VM level.
        defaulted_transaction = transaction.copy()
        if "gas" not in defaulted_transaction:
            defaulted_transaction["gas"] = self._max_available_gas()

        signed_evm_transaction = self._get_normalized_and_signed_evm_transaction(
            defaulted_transaction,
            block_number,
        )

        computation = _execute_and_revert_transaction(
            self.chain,
            signed_evm_transaction,
            block_number,
        )
        if computation.is_error:
            msg = str(computation._error)

            # Check to see if it's a EIP838 standard error, with ABI signature
            # of Error(string). If so - extract the message/reason.
            if self.is_eip838_error(computation._error):
                error_str = computation._error.args[0][4:]
                try:
                    decoded_args = abi.decode(["string"], error_str)
                    msg = decoded_args[0]
                except DecodingError:
                    # Invalid encoded bytes, leave msg as computation._error
                    # byte string.
                    pass

            raise TransactionFailed(msg)

        return computation.output



from __future__ import absolute_import

from .base import (
    BaseValidator,
)
from .common import (
    validate_uint256,
)
from .inbound import (
    validate_account as validate_inbound_account,
    validate_block_hash as validate_inbound_block_hash,
    validate_block_number as validate_inbound_block_number,
    validate_filter_id as validate_inbound_filter_id,
    validate_filter_params as validate_inbound_filter_params,
    validate_private_key as validate_inbound_private_key,
    validate_raw_transaction as validate_inbound_raw_transaction,
    validate_inbound_storage_slot as validate_inbound_storage_slot,
    validate_timestamp as validate_inbound_timestamp,
    validate_transaction as validate_inbound_transaction,
    validate_transaction_hash as validate_inbound_transaction_hash,
)
from .outbound import (
    validate_32_byte_string,
    validate_accounts as validate_outbound_accounts,
    validate_block as validate_outbound_block,
    validate_block_hash as validate_outbound_block_hash,
    validate_bytes as validate_outbound_bytes,
    validate_log_entry as validate_outbound_log_entry,
    validate_receipt as validate_outbound_receipt,
    validate_transaction as validate_outbound_transaction,
)


class DefaultValidator(BaseValidator):
    #
    # Inbound
    #
    validate_inbound_account = staticmethod(validate_inbound_account)
    validate_inbound_block_hash = staticmethod(validate_inbound_block_hash)
    validate_inbound_block_number = staticmethod(validate_inbound_block_number)
    validate_inbound_filter_id = staticmethod(validate_inbound_filter_id)
    validate_inbound_filter_params = staticmethod(validate_inbound_filter_params)
    validate_inbound_private_key = staticmethod(validate_inbound_private_key)
    validate_inbound_raw_transaction = staticmethod(validate_inbound_raw_transaction)
    validate_inbound_storage_slot = staticmethod(validate_inbound_storage_slot)
    validate_inbound_timestamp = staticmethod(validate_inbound_timestamp)
    validate_inbound_transaction = staticmethod(validate_inbound_transaction)
    validate_inbound_transaction_hash = staticmethod(validate_inbound_transaction_hash)

    #
    # Outbound
    #
    validate_outbound_accounts = staticmethod(validate_outbound_accounts)
    validate_outbound_balance = staticmethod(validate_uint256)
    validate_outbound_block = staticmethod(validate_outbound_block)
    validate_outbound_block_hash = staticmethod(validate_outbound_block_hash)
    validate_outbound_code = staticmethod(validate_outbound_bytes)
    validate_outbound_gas_estimate = staticmethod(validate_uint256)
    validate_outbound_nonce = staticmethod(validate_uint256)
    validate_outbound_log_entry = staticmethod(validate_outbound_log_entry)
    validate_outbound_receipt = staticmethod(validate_outbound_receipt)
    validate_outbound_return_data = staticmethod(validate_outbound_bytes)
    validate_outbound_storage = staticmethod(validate_uint256)
    validate_outbound_transaction = staticmethod(validate_outbound_transaction)
    validate_outbound_transaction_hash = staticmethod(validate_32_byte_string)



from eth_utils import to_int

import rlp

from toolz import (
    merge,
)

from .utils import is_supported_pyevm_version_available


if is_supported_pyevm_version_available():
    from eth.rlp.transactions import BaseTransaction
    from eth.vm.forks.berlin.transactions import TypedTransaction
else:
    BaseTransaction = None
    TypedTransaction = None

from eth_tester.exceptions import ValidationError
from eth_tester.utils.address import (
    generate_contract_address,
)
from eth_tester.utils.encoding import (
    int_to_32byte_big_endian,
)


def pad32(value):
    return value.rjust(32, b"\x00")


def serialize_block(block, full_transaction, is_pending):
    if full_transaction:
        transaction_serializer = serialize_transaction
    else:
        transaction_serializer = serialize_transaction_hash

    transactions = [
        transaction_serializer(block, transaction, index, is_pending)
        for index, transaction in enumerate(block.transactions)
    ]

    if block.uncles:
        raise NotImplementedError("Uncle serialization has not been implemented")

    block_info = {
        "number": block.header.block_number,
        "hash": block.header.hash,
        "parent_hash": block.header.parent_hash,
        "nonce": block.header.nonce,
        "sha3_uncles": block.header.uncles_hash,
        "logs_bloom": block.header.bloom,
        "transactions_root": block.header.transaction_root,
        "receipts_root": block.header.receipt_root,
        "state_root": block.header.state_root,
        "coinbase": block.header.coinbase,
        "difficulty": block.header.difficulty,
        "total_difficulty": block.header.difficulty,  # TODO: actual total difficulty
        "mix_hash": block.header.mix_hash,
        "size": len(rlp.encode(block)),
        "extra_data": pad32(block.header.extra_data),
        "gas_limit": block.header.gas_limit,
        "gas_used": block.header.gas_used,
        "timestamp": block.header.timestamp,
        "transactions": transactions,
        "uncles": [uncle.hash for uncle in block.uncles],
    }

    if hasattr(block.header, "base_fee_per_gas"):
        base_fee = block.header.base_fee_per_gas
        block_info.update({"base_fee_per_gas": base_fee})

    if hasattr(block.header, "withdrawals_root") and hasattr(block, "withdrawals"):
        block_info.update({"withdrawals": serialize_block_withdrawals(block)})
        block_info.update({"withdrawals_root": block.header.withdrawals_root})

    return block_info


def serialize_transaction_hash(block, transaction, transaction_index, is_pending):
    return transaction.hash


def serialize_transaction(block, transaction, transaction_index, is_pending):
    txn_type = _extract_transaction_type(transaction)

    common_transaction_params = {
        "type": txn_type,
        "hash": transaction.hash,
        "nonce": transaction.nonce,
        "block_hash": None if is_pending else block.hash,
        "block_number": None if is_pending else block.number,
        "transaction_index": None if is_pending else transaction_index,
        "from": transaction.sender,
        "to": transaction.to,
        "value": transaction.value,
        "gas": transaction.gas,
        "data": transaction.data,
        "r": transaction.r,
        "s": transaction.s,
        "v": transaction.v
        if _field_in_transaction(transaction, "v")
        else transaction.y_parity,
    }
    if _field_in_transaction(transaction, "gas_price"):
        type_specific_params = {"gas_price": transaction.gas_price}

        if _field_in_transaction(transaction, "access_list"):
            # access list transaction
            type_specific_params = merge(
                type_specific_params,
                {
                    "chain_id": transaction.chain_id,
                    "access_list": transaction.access_list or (),
                },
            )
    elif any(
        _field_in_transaction(transaction, _)
        for _ in ("max_fee_per_gas" and "max_priority_fee_per_gas")
    ):
        # dynamic fee transaction
        type_specific_params = {
            "chain_id": transaction.chain_id,
            "max_fee_per_gas": transaction.max_fee_per_gas,
            "max_priority_fee_per_gas": transaction.max_priority_fee_per_gas,
            "access_list": transaction.access_list or (),
            # TODO: Sometime in 2022 the inclusion of gas_price may be removed from
            #  dynamic fee transactions and we can get rid of this behavior.
            #  https://github.com/ethereum/execution-specs/pull/251
            "gas_price": (
                transaction.max_fee_per_gas
                if is_pending
                else _calculate_effective_gas_price(transaction, block, txn_type)
            ),
        }
    else:
        raise ValidationError("Invariant: code path should be unreachable")

    return merge(common_transaction_params, type_specific_params)


def _field_in_transaction(transaction, field):
    """
    There are many classes of transactions, we have to be able to search for a
    particular field depending on the type of transaction - from dict, to legacy
    transaction classes, to *TypedTransaction classes.
    """
    if isinstance(transaction, dict):
        return field in transaction
    elif isinstance(transaction, BaseTransaction):
        # all legacy transactions inherit from BaseTransaction
        return field in transaction.as_dict()
    elif isinstance(transaction, TypedTransaction):
        # all typed transactions inherit from TypedTransaction
        return hasattr(transaction, field)


def serialize_transaction_receipt(
    block, receipts, transaction, transaction_index, is_pending
):
    receipt = receipts[transaction_index]
    _txn_type = _extract_transaction_type(transaction)
    state_root = receipt.state_root

    if transaction.to == b"":
        contract_addr = generate_contract_address(
            transaction.sender,
            transaction.nonce,
        )
    else:
        contract_addr = None

    if transaction_index == 0:
        origin_gas = 0
    else:
        origin_gas = receipts[transaction_index - 1].gas_used

    return {
        "block_hash": None if is_pending else block.hash,
        "block_number": None if is_pending else block.number,
        "contract_address": contract_addr,
        "cumulative_gas_used": receipt.gas_used,
        "effective_gas_price": _calculate_effective_gas_price(
            transaction, block, _txn_type
        ),
        "from": transaction.sender,
        "gas_used": receipt.gas_used - origin_gas,
        "logs": [
            serialize_log(
                block, transaction, transaction_index, log, log_index, is_pending
            )
            for log_index, log in enumerate(receipt.logs)
        ],
        "state_root": state_root,
        "status": to_int(state_root),
        "to": transaction.to,
        "transaction_hash": transaction.hash,
        "transaction_index": None if is_pending else transaction_index,
        "type": _txn_type,
    }


def serialize_log(block, transaction, transaction_index, log, log_index, is_pending):
    return {
        "type": "pending" if is_pending else "mined",
        "log_index": log_index,
        "transaction_index": None if is_pending else transaction_index,
        "transaction_hash": transaction.hash,
        "block_hash": None if is_pending else block.hash,
        "block_number": None if is_pending else block.number,
        "address": log.address,
        "data": log.data,
        "topics": [int_to_32byte_big_endian(topic) for topic in log.topics],
    }


def _extract_transaction_type(transaction):
    if isinstance(transaction, TypedTransaction):
        try:
            transaction.gas_price  # noqa: 201
            return "0x1"
        except AttributeError:
            return "0x2"
    # legacy transactions being '0x0' taken from current geth version v1.10.10
    return "0x0"


def _calculate_effective_gas_price(transaction, block, transaction_type):
    return (
        min(
            transaction.max_fee_per_gas,
            transaction.max_priority_fee_per_gas + block.header.base_fee_per_gas,
        )
        if transaction_type == "0x2"
        else transaction.gas_price
    )


def serialize_block_withdrawals(block):
    return [
        {
            "index": withdrawal.index,
            "validator_index": withdrawal.validator_index,
            "address": withdrawal.address,
            "amount": withdrawal.amount,
        }
        for withdrawal in block.withdrawals
    ]



from eth_utils import to_bytes

from eth_tester.backends.mock.common import (
    calculate_effective_gas_price,
)
from eth_tester.utils.transactions import (
    extract_transaction_type,
)
from eth_utils.toolz import (
    assoc,
    partial,
    pipe,
)


def serialize_block(block, transaction_serializer, is_pending):
    serialized_transactions = tuple(
        transaction_serializer(
            transaction, block, transaction_index, is_pending=is_pending
        )
        for transaction_index, transaction in enumerate(block["transactions"])
    )
    block_with_transactions = assoc(block, "transactions", serialized_transactions)
    block_with_withdrawals = assoc(
        block_with_transactions,
        "withdrawals",
        block["withdrawals"],
    )
    return block_with_withdrawals


def serialize_transaction_as_hash(transaction, block, transaction_index, is_pending):
    return transaction["hash"]


def serialize_full_transaction(transaction, block, transaction_index, is_pending):
    if is_pending:
        block_number = None
        block_hash = None
        transaction_index = None
    else:
        block_number = block["number"]
        block_hash = block["hash"]

    serialized_transaction = pipe(
        transaction,
        partial(assoc, key="block_number", value=block_number),
        partial(assoc, key="block_hash", value=block_hash),
        partial(assoc, key="transaction_index", value=transaction_index),
        partial(assoc, key="type", value=extract_transaction_type(transaction)),
    )

    if "gas_price" in transaction:
        return serialized_transaction
    else:
        # TODO: Sometime in 2022 the inclusion of gas_price may be removed from
        #  dynamic fee transactions and we can get rid of this behavior.
        #  https://github.com/ethereum/execution-specs/pull/251
        gas_price = (
            transaction["max_fee_per_gas"]
            if is_pending
            else calculate_effective_gas_price(transaction, block)
        )
        return assoc(serialized_transaction, "gas_price", gas_price)


def serialize_receipt(receipt, transaction, block, transaction_index, is_pending):
    if is_pending:
        block_number = None
        block_hash = None
        transaction_index = None
    else:
        block_number = block["number"]
        block_hash = block["hash"]

    return pipe(
        receipt,
        partial(assoc, key="block_number", value=block_number),
        partial(assoc, key="block_hash", value=block_hash),
        partial(
            assoc,
            key="effective_gas_price",
            value=(calculate_effective_gas_price(transaction, block)),
        ),
        partial(assoc, key="from", value=to_bytes(transaction["from"])),
        partial(assoc, key="state_root", value=b"\x00"),
        partial(assoc, key="status", value=0),
        partial(assoc, key="to", value=transaction["to"]),
        partial(assoc, key="transaction_index", value=transaction_index),
        partial(assoc, key="type", value=extract_transaction_type(transaction)),
    )



import pkg_resources
import sys

from .main import (  # noqa: F401
    EthereumTester,
)

from .backends import (  # noqa: F401
    MockBackend,
    PyEVMBackend,
)


if sys.version_info.major < 3:
    raise EnvironmentError("eth-tester only supports Python 3")


__version__ = pkg_resources.get_distribution("eth-tester").version



def calculate_effective_gas_price(transaction, block):
    return (
        min(
            transaction["max_fee_per_gas"],
            transaction["max_priority_fee_per_gas"] + block["base_fee_per_gas"],
        )
        if "max_fee_per_gas" in transaction
        else transaction["gas_price"]
    )



from __future__ import unicode_literals

import collections
import itertools
import operator
import time
import functools

from eth_typing import (
    HexAddress,
    HexStr,
)
from eth_utils import (
    is_integer,
    is_same_address,
    to_list,
    to_tuple,
)

from eth_utils.toolz import (
    assoc,
    dissoc,
    remove,
    compose,
    excepts,
    partial,
)

from eth_tester.backends import (
    get_chain_backend,
)
from eth_tester.constants import (
    ZERO_ADDRESS_HEX,
)
from eth_tester.exceptions import (
    AccountLocked,
    BlockNotFound,
    FilterNotFound,
    SnapshotNotFound,
    TransactionNotFound,
    ValidationError,
)
from eth_tester.normalization import (
    get_normalizer_backend,
)
from eth_tester.utils.accounts import (
    private_key_to_address,
)
from eth_tester.utils.filters import (
    Filter,
    check_if_log_matches,
)
from eth_tester.utils.transactions import (
    extract_transaction_type,
    extract_valid_transaction_params,
    remove_matching_transaction_from_list,
)
from eth_tester.validation import (
    get_validator,
)


def backend_proxy_method(backend_method_name):
    def proxy_method(self, *args, **kwargs):
        backend_method = getattr(self.backend, backend_method_name)
        return backend_method(*args, **kwargs)

    return proxy_method


def handle_auto_mining(func):
    @functools.wraps(func)
    def func_wrapper(self, *args, **kwargs):
        if self.auto_mine_transactions:
            transaction_hash = func(self, *args, **kwargs)
            self.mine_block()
        else:
            snapshot = self.take_snapshot()
            try:
                transaction_hash = func(self, *args, **kwargs)
                pending_transaction = self.get_transaction_by_hash(transaction_hash)
                # Remove any pending transactions with the same nonce
                self._pending_transactions = remove_matching_transaction_from_list(
                    self._pending_transactions, pending_transaction
                )
                cleaned_transaction = _clean_pending_transaction(pending_transaction)
                self._pending_transactions.append(cleaned_transaction)
            finally:
                self.revert_to_snapshot(snapshot)
        return transaction_hash

    def _clean_pending_transaction(pending_transaction):
        cleaned_transaction = dissoc(pending_transaction, "type")

        # TODO: Sometime in 2022 the inclusion of gas_price may be removed from
        #  dynamic fee transactions and we can get rid of this behavior.
        #  https://github.com/ethereum/execution-specs/pull/251
        # remove gas_price for dynamic fee transactions
        if "gas_price" and "max_fee_per_gas" in pending_transaction:
            cleaned_transaction = dissoc(cleaned_transaction, "gas_price")

        return cleaned_transaction

    return func_wrapper


class EthereumTester:
    backend = None

    validator = None
    normalizer = None

    fork_blocks = None

    auto_mine_transactions = None

    def __init__(
        self, backend=None, validator=None, normalizer=None, auto_mine_transactions=True
    ):
        if backend is None:
            backend = get_chain_backend()

        if validator is None:
            validator = get_validator()

        if normalizer is None:
            normalizer = get_normalizer_backend()

        self.backend = backend
        self.validator = validator
        self.normalizer = normalizer

        self.auto_mine_transactions = auto_mine_transactions

        self._reset_local_state()

    #
    # Private API
    #
    _filter_counter = None
    _log_filters = None
    _block_filters = None
    _pending_transaction_filters = None

    _pending_transactions = []

    _snapshot_counter = None
    _snapshots = None

    _account_passwords = None
    _account_unlock = None

    def _reset_local_state(self):
        # filter tracking
        self._filter_counter = itertools.count()
        self._log_filters = {}
        self._block_filters = {}
        self._pending_transaction_filters = {}

        # snapshot tracking
        self._snapshot_counter = itertools.count()
        self._snapshots = {}

        # raw accounts
        self._account_passwords = {}
        self._account_unlock = collections.defaultdict(lambda: False)

    #
    # Time Traveling
    #
    def time_travel(self, to_timestamp):
        self.validator.validate_inbound_timestamp(to_timestamp)
        # make sure we are not traveling back in time as this is not possible.
        current_timestamp = self.get_block_by_number("pending")["timestamp"]
        if to_timestamp == current_timestamp:
            # no change, return immediately
            return
        elif to_timestamp < current_timestamp:
            raise ValidationError(
                "Space time continuum distortion detected.  Traveling backwards "
                "in time violates interdimensional ordinance 31415-926."
            )
        else:
            self.backend.time_travel(to_timestamp)

    #
    # Accounts
    #
    def get_accounts(self):
        raw_accounts = self.backend.get_accounts()
        self.validator.validate_outbound_accounts(raw_accounts)
        accounts = self.normalizer.normalize_outbound_account_list(raw_accounts)
        return accounts

    def add_account(self, private_key, password=None):
        # TODO: validation
        self.validator.validate_inbound_private_key(private_key)
        raw_private_key = self.normalizer.normalize_inbound_private_key(private_key)
        raw_account = private_key_to_address(raw_private_key)
        account = self.normalizer.normalize_outbound_account(raw_account)
        if any(is_same_address(account, value) for value in self.get_accounts()):
            raise ValidationError("Account already present in account list")

        self.backend.add_account(raw_private_key)
        self._account_passwords[raw_account] = password
        # TODO: outbound normalization
        return account

    def unlock_account(self, account, password, unlock_seconds=None):
        self.validator.validate_inbound_account(account)
        raw_account = self.normalizer.normalize_inbound_account(account)
        try:
            account_password = self._account_passwords[raw_account]
        except KeyError:
            raise ValidationError("Unknown account")

        if account_password is None:
            raise ValidationError("Account does not have a password")

        if account_password != password:
            raise ValidationError("Wrong password")

        if unlock_seconds is None:
            unlock_until = None
        else:
            unlock_until = time.time() + unlock_seconds

        self._account_unlock[raw_account] = unlock_until

    def lock_account(self, account):
        self.validator.validate_inbound_account(account)
        raw_account = self.normalizer.normalize_inbound_account(account)

        if raw_account not in self._account_passwords:
            raise ValidationError("Unknown account")
        elif self._account_passwords[raw_account] is None:
            raise ValidationError("Account does not have a password")

        self._account_unlock[raw_account] = False

    def get_balance(self, account, block_number="latest"):
        self.validator.validate_inbound_account(account)
        self.validator.validate_inbound_block_number(block_number)
        raw_account = self.normalizer.normalize_inbound_account(account)
        raw_block_number = self.normalizer.normalize_inbound_block_number(block_number)
        raw_balance = self.backend.get_balance(raw_account, raw_block_number)
        self.validator.validate_outbound_balance(raw_balance)
        balance = self.normalizer.normalize_outbound_balance(raw_balance)
        return balance

    def get_code(self, account, block_number="latest"):
        self.validator.validate_inbound_account(account)
        self.validator.validate_inbound_block_number(block_number)
        raw_account = self.normalizer.normalize_inbound_account(account)
        raw_block_number = self.normalizer.normalize_inbound_block_number(block_number)
        raw_code = self.backend.get_code(raw_account, raw_block_number)
        self.validator.validate_outbound_code(raw_code)
        code = self.normalizer.normalize_outbound_code(raw_code)
        return code

    def get_storage_at(
        self,
        account: HexAddress,
        slot: HexStr,
        # properly type hint once eth-typing brings in updated `BlockIdentifier`
        block_number="latest",
    ) -> int:
        self.validator.validate_inbound_account(account)
        self.validator.validate_inbound_storage_slot(slot)
        self.validator.validate_inbound_block_number(block_number)
        raw_account = self.normalizer.normalize_inbound_account(account)
        raw_slot = self.normalizer.normalize_inbound_storage_slot(slot)
        raw_block_number = self.normalizer.normalize_inbound_block_number(block_number)
        raw_storage = self.backend.get_storage(raw_account, raw_slot, raw_block_number)
        self.validator.validate_outbound_storage(raw_storage)
        storage = self.normalizer.normalize_outbound_storage(raw_storage)
        return storage

    def get_nonce(self, account, block_number="latest"):
        self.validator.validate_inbound_account(account)
        self.validator.validate_inbound_block_number(block_number)
        raw_account = self.normalizer.normalize_inbound_account(account)
        raw_block_number = self.normalizer.normalize_inbound_block_number(block_number)
        raw_nonce = self.backend.get_nonce(raw_account, raw_block_number)
        self.validator.validate_outbound_nonce(raw_nonce)
        nonce = self.normalizer.normalize_outbound_nonce(raw_nonce)
        return nonce

    #
    # Blocks, Transactions, Receipts
    #

    @staticmethod
    def _normalize_pending_transaction(pending_transaction):
        """
        Add the transaction type and, if a dynamic fee transaction, add gas_price =
        max_fee_per_gas as highlighted in the execution-specs link below.
        """
        _type = extract_transaction_type(pending_transaction)
        pending_transaction = assoc(pending_transaction, "type", _type)

        # TODO: Sometime in 2022 the inclusion of gas_price may be removed from
        # dynamic fee transactions and we can get rid of this behavior.
        # https://github.com/ethereum/execution-specs/pull/251
        # add gas_price = max_fee_per_gas to pending dynamic fee transactions
        if _type == "0x2":
            pending_transaction = assoc(
                pending_transaction, "gas_price", pending_transaction["max_fee_per_gas"]
            )
        return pending_transaction

    def _get_pending_transaction_by_hash(self, transaction_hash):
        for transaction in self._pending_transactions:
            if transaction["hash"] == transaction_hash:
                return transaction
        raise TransactionNotFound(
            f"No transaction found for transaction hash: {transaction_hash}"
        )

    def get_transaction_by_hash(self, transaction_hash):
        self.validator.validate_inbound_transaction_hash(transaction_hash)
        try:
            pending_transaction = self._get_pending_transaction_by_hash(
                transaction_hash
            )
            return self._normalize_pending_transaction(pending_transaction)
        except TransactionNotFound:
            raw_transaction_hash = self.normalizer.normalize_inbound_transaction_hash(
                transaction_hash,
            )
            raw_transaction = self.backend.get_transaction_by_hash(raw_transaction_hash)
            self.validator.validate_outbound_transaction(raw_transaction)
            transaction = self.normalizer.normalize_outbound_transaction(
                raw_transaction
            )
            return transaction

    def get_block_by_number(self, block_number="latest", full_transactions=False):
        self.validator.validate_inbound_block_number(block_number)
        raw_block_number = self.normalizer.normalize_inbound_block_number(block_number)
        raw_block = self.backend.get_block_by_number(
            raw_block_number, full_transactions
        )
        self.validator.validate_outbound_block(raw_block)
        block = self.normalizer.normalize_outbound_block(raw_block)
        return block

    def get_block_by_hash(self, block_hash, full_transactions=False):
        self.validator.validate_inbound_block_hash(block_hash)
        raw_block_hash = self.normalizer.normalize_inbound_block_hash(block_hash)
        raw_block = self.backend.get_block_by_hash(raw_block_hash, full_transactions)
        self.validator.validate_outbound_block(raw_block)
        block = self.normalizer.normalize_outbound_block(raw_block)
        return block

    def get_transaction_receipt(self, transaction_hash):
        self.validator.validate_inbound_transaction_hash(transaction_hash)
        raw_transaction_hash = self.normalizer.normalize_inbound_transaction_hash(
            transaction_hash,
        )
        raw_receipt = self.backend.get_transaction_receipt(raw_transaction_hash)
        self.validator.validate_outbound_receipt(raw_receipt)
        receipt = self.normalizer.normalize_outbound_receipt(raw_receipt)
        return receipt

    #
    # Mining
    #
    def enable_auto_mine_transactions(self):
        self.auto_mine_transactions = True
        sent_transaction_hashes = self._pop_pending_transactions_to_pending_block()
        self.mine_block()
        return sent_transaction_hashes

    def disable_auto_mine_transactions(self):
        self.auto_mine_transactions = False

    def mine_blocks(self, num_blocks=1, coinbase=ZERO_ADDRESS_HEX):
        self.validator.validate_inbound_account(coinbase)
        normalized_coinbase = self.normalizer.normalize_inbound_account(coinbase)

        if not self.auto_mine_transactions:
            self._pop_pending_transactions_to_pending_block()

        raw_block_hashes = self.backend.mine_blocks(num_blocks, normalized_coinbase)

        if len(raw_block_hashes) != num_blocks:
            raise ValidationError(
                f"Invariant: tried to mine {num_blocks} blocks.  Got "
                f"{len(raw_block_hashes)} mined block hashes."
            )

        for raw_block_hash in raw_block_hashes:
            self.validator.validate_outbound_block_hash(raw_block_hash)
        block_hashes = [
            self.normalizer.normalize_outbound_block_hash(raw_block_hash)
            for raw_block_hash in raw_block_hashes
        ]

        # feed the block hashes to any block filters
        for block_hash in block_hashes:
            block = self.get_block_by_hash(block_hash)

            for _, block_filter in self._block_filters.items():
                raw_block_hash = self.normalizer.normalize_inbound_block_hash(
                    block_hash
                )
                block_filter.add(raw_block_hash)

            self._process_block_logs(block)

        return block_hashes

    def mine_block(self, coinbase=ZERO_ADDRESS_HEX):
        self.validator.validate_inbound_account(coinbase)
        block_hash = self.mine_blocks(1, coinbase=coinbase)[0]
        return block_hash

    #
    # Private mining API
    #
    def _process_block_logs(self, block):
        for fid, filter in self._log_filters.items():
            self._add_log_entries_to_filter(block, filter)

    def _add_log_entries_to_filter(self, block, filter_):
        for transaction_hash in block["transactions"]:
            receipt = self.get_transaction_receipt(transaction_hash)
            for log_entry in receipt["logs"]:
                raw_log_entry = self.normalizer.normalize_inbound_log_entry(log_entry)
                filter_.add(raw_log_entry)

    def _pop_pending_transactions_to_pending_block(self):
        sent_transaction_hashes = self._add_all_to_pending_block(
            self._pending_transactions
        )
        self._pending_transactions.clear()
        return sent_transaction_hashes

    @to_list
    def _add_all_to_pending_block(self, pending_transactions):
        for pending in pending_transactions:
            txn = extract_valid_transaction_params(pending)
            yield self._add_transaction_to_pending_block(
                txn, txn_internal_type="send_signed"
            )

    #
    # Transaction Sending
    #
    def _handle_filtering_for_transaction(self, transaction_hash):
        # feed the transaction hash to any pending transaction filters.
        for _, filter in self._pending_transaction_filters.items():
            raw_transaction_hash = self.normalizer.normalize_inbound_transaction_hash(
                transaction_hash,
            )
            filter.add(raw_transaction_hash)

        if self._log_filters:
            receipt = self.get_transaction_receipt(transaction_hash)
            for log_entry in receipt["logs"]:
                for _, filter in self._log_filters.items():
                    raw_log_entry = self.normalizer.normalize_inbound_log_entry(
                        log_entry
                    )
                    filter.add(raw_log_entry)

    @handle_auto_mining
    def send_raw_transaction(self, raw_transaction_hex):
        self.validator.validate_inbound_raw_transaction(raw_transaction_hex)
        raw_transaction = self.normalizer.normalize_inbound_raw_transaction(
            raw_transaction_hex
        )
        raw_transaction_hash = self.backend.send_raw_transaction(raw_transaction)
        self.validator.validate_outbound_transaction_hash(raw_transaction_hash)
        transaction_hash = self.normalizer.normalize_outbound_transaction_hash(
            raw_transaction_hash,
        )
        self._handle_filtering_for_transaction(transaction_hash)
        return transaction_hash

    @handle_auto_mining
    def send_transaction(self, transaction):
        return self._add_transaction_to_pending_block(transaction)

    def call(self, transaction, block_number="latest"):
        self.validator.validate_inbound_transaction(
            transaction, txn_internal_type="call"
        )
        raw_transaction = self.normalizer.normalize_inbound_transaction(transaction)
        self.validator.validate_inbound_block_number(block_number)
        raw_block_number = self.normalizer.normalize_inbound_block_number(block_number)
        raw_result = self.backend.call(raw_transaction, raw_block_number)
        self.validator.validate_outbound_return_data(raw_result)
        result = self.normalizer.normalize_outbound_return_data(raw_result)
        return result

    def estimate_gas(self, transaction, block_number="latest"):
        self.validator.validate_inbound_transaction(
            transaction, txn_internal_type="estimate"
        )
        raw_transaction = self.normalizer.normalize_inbound_transaction(transaction)
        self.validator.validate_inbound_block_number(block_number)
        raw_block_number = self.normalizer.normalize_inbound_block_number(block_number)
        raw_gas_estimate = self.backend.estimate_gas(raw_transaction, raw_block_number)
        self.validator.validate_outbound_gas_estimate(raw_gas_estimate)
        gas_estimate = self.normalizer.normalize_outbound_gas_estimate(raw_gas_estimate)
        return gas_estimate

    #
    # Private Transaction API
    #
    def _add_transaction_to_pending_block(self, transaction, txn_internal_type="send"):
        self.validator.validate_inbound_transaction(
            transaction, txn_internal_type=txn_internal_type
        )
        raw_transaction = self.normalizer.normalize_inbound_transaction(transaction)

        if raw_transaction["from"] in self._account_passwords:
            unlocked_until = self._account_unlock[raw_transaction["from"]]
            account_password = self._account_passwords[raw_transaction["from"]]
            is_locked = (
                account_password is not None
                and unlocked_until is not None
                and (unlocked_until is False or time.time() > unlocked_until)
            )
            if is_locked:
                raise AccountLocked("The account is currently locked")

        if {"r", "s", "v"}.issubset(transaction.keys()):
            try:
                raw_transaction_hash = self.backend.send_signed_transaction(
                    raw_transaction
                )
            except NotImplementedError:
                unsigned_transaction = dissoc(raw_transaction, "r", "s", "v")
                raw_transaction_hash = self.backend.send_transaction(
                    unsigned_transaction
                )
        else:
            raw_transaction_hash = self.backend.send_transaction(raw_transaction)

        self.validator.validate_outbound_transaction_hash(raw_transaction_hash)
        transaction_hash = self.normalizer.normalize_outbound_transaction_hash(
            raw_transaction_hash,
        )

        self._handle_filtering_for_transaction(transaction_hash)

        return transaction_hash

    #
    # Snapshot and Revert
    #
    def take_snapshot(self):
        snapshot = self.backend.take_snapshot()
        snapshot_id = next(self._snapshot_counter)
        self._snapshots[snapshot_id] = snapshot
        return snapshot_id

    def revert_to_snapshot(self, snapshot_id):
        try:
            snapshot = self._snapshots[snapshot_id]
        except KeyError:
            raise SnapshotNotFound(f"No snapshot found for id: {snapshot_id}")
        else:
            self.backend.revert_to_snapshot(snapshot)

        for block_filter in self._block_filters.values():
            self._revert_block_filter(block_filter)
        for pending_transaction_filter in self._pending_transaction_filters.values():
            self._revert_pending_transaction_filter(pending_transaction_filter)
        for log_filter in self._log_filters.values():
            self._revert_log_filter(log_filter)

    def reset_to_genesis(self):
        self.backend.reset_to_genesis()
        self._reset_local_state()

    #
    # Private filter API
    #
    def _revert_block_filter(self, filter):
        is_valid_block_hash = excepts(
            (BlockNotFound,),
            compose(
                bool,
                self.get_block_by_hash,
                self.normalizer.normalize_outbound_block_hash,
            ),
            lambda v: False,
        )
        values_to_remove = tuple(remove(is_valid_block_hash, filter.get_all()))
        filter.remove(*values_to_remove)

    def _revert_pending_transaction_filter(self, filter):
        is_valid_transaction_hash = excepts(
            (TransactionNotFound,),
            compose(
                bool,
                self.get_transaction_by_hash,
                self.normalizer.normalize_outbound_transaction_hash,
            ),
            lambda v: False,
        )
        values_to_remove = remove(is_valid_transaction_hash, filter.get_all())
        filter.remove(*values_to_remove)

    def _revert_log_filter(self, filter):
        is_valid_transaction_hash = excepts(
            (TransactionNotFound,),
            compose(
                bool,
                self.get_transaction_by_hash,
                self.normalizer.normalize_outbound_transaction_hash,
                operator.itemgetter("transaction_hash"),
            ),
            lambda v: False,
        )
        values_to_remove = remove(is_valid_transaction_hash, filter.get_all())
        filter.remove(*values_to_remove)

    #
    # Filters
    #
    def create_block_filter(self):
        raw_filter_id = next(self._filter_counter)
        self._block_filters[raw_filter_id] = Filter(filter_params=None)
        filter_id = self.normalizer.normalize_outbound_filter_id(raw_filter_id)
        return filter_id

    def create_pending_transaction_filter(self):
        raw_filter_id = next(self._filter_counter)
        self._pending_transaction_filters[raw_filter_id] = Filter(filter_params=None)
        filter_id = self.normalizer.normalize_outbound_filter_id(raw_filter_id)
        return filter_id

    def create_log_filter(
        self, from_block=None, to_block=None, address=None, topics=None
    ):
        self.validator.validate_inbound_filter_params(
            from_block=from_block,
            to_block=to_block,
            address=address,
            topics=topics,
        )
        (
            raw_from_block,
            raw_to_block,
            raw_address,
            raw_topics,
        ) = self.normalizer.normalize_inbound_filter_params(
            from_block=from_block,
            to_block=to_block,
            address=address,
            topics=topics,
        )

        raw_filter_id = next(self._filter_counter)
        raw_filter_params = {
            "from_block": raw_from_block,
            "to_block": raw_to_block,
            "addresses": raw_address,
            "topics": raw_topics,
        }
        filter_fn = partial(check_if_log_matches, **raw_filter_params)
        new_filter = Filter(
            filter_params=raw_filter_params,
            filter_fn=filter_fn,
        )
        self._log_filters[raw_filter_id] = new_filter

        if is_integer(raw_from_block):
            if is_integer(raw_to_block):
                upper_bound = raw_to_block + 1
            else:
                upper_bound = self.get_block_by_number("pending")["number"]
            for block_number in range(raw_from_block, upper_bound):
                block = self.get_block_by_number(block_number)
                self._add_log_entries_to_filter(block, new_filter)

        filter_id = self.normalizer.normalize_outbound_filter_id(raw_filter_id)
        return filter_id

    def delete_filter(self, filter_id):
        self.validator.validate_inbound_filter_id(filter_id)
        raw_filter_id = self.normalizer.normalize_inbound_filter_id(filter_id)

        if raw_filter_id in self._block_filters:
            del self._block_filters[raw_filter_id]
        elif raw_filter_id in self._pending_transaction_filters:
            del self._pending_transaction_filters[raw_filter_id]
        elif raw_filter_id in self._log_filters:
            del self._log_filters[raw_filter_id]
        else:
            raise FilterNotFound("Unknown filter id")

    @to_tuple
    def get_only_filter_changes(self, filter_id):
        self.validator.validate_inbound_filter_id(filter_id)
        raw_filter_id = self.normalizer.normalize_inbound_filter_id(filter_id)

        if raw_filter_id in self._block_filters:
            filter = self._block_filters[raw_filter_id]
            normalize_fn = self.normalizer.normalize_outbound_block_hash
        elif raw_filter_id in self._pending_transaction_filters:
            filter = self._pending_transaction_filters[raw_filter_id]
            normalize_fn = self.normalizer.normalize_outbound_transaction_hash
        elif raw_filter_id in self._log_filters:
            filter = self._log_filters[raw_filter_id]
            normalize_fn = self.normalizer.normalize_outbound_log_entry
        else:
            raise FilterNotFound("Unknown filter id")

        for item in filter.get_changes():
            yield normalize_fn(item)

    @to_tuple
    def get_all_filter_logs(self, filter_id):
        self.validator.validate_inbound_filter_id(filter_id)
        raw_filter_id = self.normalizer.normalize_inbound_filter_id(filter_id)

        if raw_filter_id in self._block_filters:
            filter = self._block_filters[raw_filter_id]
            normalize_fn = self.normalizer.normalize_outbound_block_hash
        elif raw_filter_id in self._pending_transaction_filters:
            filter = self._pending_transaction_filters[raw_filter_id]
            normalize_fn = self.normalizer.normalize_outbound_transaction_hash
        elif raw_filter_id in self._log_filters:
            filter = self._log_filters[raw_filter_id]
            normalize_fn = self.normalizer.normalize_outbound_log_entry
        else:
            raise FilterNotFound("Unknown filter id")

        for item in filter.get_all():
            yield normalize_fn(item)

    @to_tuple
    def get_logs(self, from_block=None, to_block=None, address=None, topics=None):
        self.validator.validate_inbound_filter_params(
            from_block=from_block,
            to_block=to_block,
            address=address,
            topics=topics,
        )
        (
            raw_from_block,
            raw_to_block,
            raw_address,
            raw_topics,
        ) = self.normalizer.normalize_inbound_filter_params(
            from_block=from_block,
            to_block=to_block,
            address=address,
            topics=topics,
        )

        # Setup the filter object
        raw_filter_params = {
            "from_block": raw_from_block,
            "to_block": raw_to_block,
            "addresses": raw_address,
            "topics": raw_topics,
        }
        filter_fn = partial(
            check_if_log_matches,
            **raw_filter_params,
        )
        log_filter = Filter(
            filter_params=raw_filter_params,
            filter_fn=filter_fn,
        )

        # Set from/to block defaults
        if raw_from_block is None:
            raw_from_block = "latest"
        if raw_to_block is None:
            raw_to_block = "latest"

        # Determine lower bound for block range.
        if isinstance(raw_from_block, int):
            lower_bound = raw_from_block
        else:
            lower_bound = self.get_block_by_number(raw_from_block)["number"]

        # Determine upper bound for block range.
        if isinstance(raw_to_block, int):
            upper_bound = raw_to_block
        else:
            upper_bound = self.get_block_by_number(raw_to_block)["number"]

        # Enumerate the blocks in the block range to find all log entries which match.
        for block_number in range(lower_bound, upper_bound + 1):
            block = self.get_block_by_number(block_number)
            for transaction_hash in block["transactions"]:
                receipt = self.get_transaction_receipt(transaction_hash)
                for log_entry in receipt["logs"]:
                    raw_log_entry = self.normalizer.normalize_inbound_log_entry(
                        log_entry
                    )
                    log_filter.add(raw_log_entry)

        # Return the matching log entries
        for item in log_filter.get_all():
            yield self.normalizer.normalize_outbound_log_entry(item)
